import ase.optimize
from ase.io import read,write
from ase.units import Bohr
from ase.constraints import FixAtoms
from ase.parallel import parprint
from mpi4py import MPI
import numpy as np
import os

"""
	Advanced usage of  GPAW calculators
	
	By default gamma-point only and LCAO basis. Per-atomic basis sets can be specified in a 
	`basis` column of samples (`None` for default basis, `szp`, `dzp`, etc), 
	otherwise default GPAW basis is used. This feature should be used with care,
	as numbers of atoms with various basis sets should be the same for all samples!
	Can be run in parallel with mpiexec and gpaw-python interpreter ...
"""

def make_basis_dict(basis_array):
	"""
		Create dictionary for ``basis`` argument of GPAW calculator from the array.
		i-th element of the ``basis_array`` corresponds to i-th atom.
	"""
	basis = {}
	for i,a in enumerate(basis_array):
		if a is None or a == "" or a.upper() == "NONE" or a.upper() == "PPW":
			pass
		else:
			basis[i] = a
	return basis

def make_gpaw_calc(a, args):
	from gpaw import GPAW, Mixer, FermiDirac
	try:
		basis = make_basis_dict(a.get_array('basis'))		
	except:
		basis = 'dzp'
	parprint ("Basis for {} = {}".format(args.label, basis), a)
	
	# https://wiki.fysik.dtu.dk/gpaw/documentation/manual.html#manual-convergence
	convergence = {'energy': 0.0005, #0.0005,  # eV / electron
					 'density': 1e-4, #1.0e-4,
					 'eigenstates': 4.0e-8,  # eV^2 / electron  This criterion does not affect LCAO calculations.
					 'bands': 'occupied',
					 'forces': float('inf')} # eV / Ang Max

	parallel={'sl_auto':True, 'domain':(2,2,1)}
	xc = 'C09-vdW' if args.xc is None else args.xc
	calc_dft = GPAW(xc=xc, symmetry={'point_group': False}, basis=basis, occupations=FermiDirac(args.smearing_width), 
		mixer=Mixer(beta=args.beta), parallel=parallel, mode='lcao', convergence=convergence, txt=args.log_dir + args.label+".gpaw.txt")
	return calc_dft

def make_qe_calc(a, args):
	from ase.calculators.espresso import Espresso
	pps = {
		#"Ti":"Ti.rel-pw91-spn-kjpaw_psl.1.0.0.UPF",
		#"C":"C.rel-pw91-n-kjpaw_psl.1.0.0.UPF",
		#"O":"O.rel-pw91-nl-kjpaw_psl.1.0.0.UPF",
		#"H":"H.rel-pw91-kjpaw_psl.1.0.0.UPF",
		
		"Ti":"Ti.pw91-spn-kjpaw_psl.1.0.0.UPF",
		"C":"C.pw91-n-kjpaw_psl.1.0.0.UPF",
		"O":"O.pw91-nl-kjpaw_psl.1.0.0.UPF",
		"H":"H.pw91-kjpaw_psl.1.0.0.UPF",
		
		
		#"Cu":"Cu.pz-dn-kjpaw_psl.0.2.UPF",
		#"Cl":"Cl.pz-n-kjpaw_psl.0.1.UPF",
		#"C":"C.rel-pz-n-rrkjus_psl.0.1.UPF",
		#"O":"O.rel-pz-n-rrkjus_psl.0.1.UPF",
		#"H":"H.rel-pz-rrkjus_psl.0.1.UPF",
		#"Ge":"Ge.rel-pz-dn-rrkjus_psl.0.2.2.UPF",
		#"Li":"Li.rel-pz-s-rrkjus_psl.0.2.1.UPF",
	}

	ecutwfc = args.Ecut / ase.units.Ry
	smearing = args.smearing_width / ase.units.Ry
	
	calc_dft = Espresso(
		#startingwfc='file', 
		#startingpot='file',
		input_dft='vdW-DF2-C09',
		pseudopotentials=pps, occupations='smearing', tstress=False, tprnfor=False, 
		ecutwfc = ecutwfc, ecutrho = ecutwfc * 4,
		smearing='m-v',
		degauss=smearing,
		lspinorb = False,
		noncolin = False,
		vdw_corr = 'grimme-d2',
		conv_thr = 1e-5,
		mixing_mode='local-TF',
		mixing_beta = args.beta,
		mixing_ndim = 4,
		adaptive_thr = True,
		scf_must_converge = True,
		#kpts=(3, 3, 1), # gamma by default 
		)
	
	return calc_dft


def calc_energy(a, args):
	
	if args.relax != 'No':
		relax = getattr(ase.optimize, args.relax)
	args.label, _ = os.path.splitext(a.filename)
	
	args.log_dir = args.samples_dir + '/.log/'
	#if MPI.COMM_WORLD.Get_rank() == 0:
	os.makedirs(args.log_dir, exist_ok=True)

	if args.calculator.upper() == 'GPAW':
		calc_dft = make_gpaw_calc(a, args)
	else:
		calc_dft = make_qe_calc(a, args)
	
	a.set_calculator(calc_dft)
	e = a.get_potential_energy()
	try:
		efc = a.get_potential_energy(force_consistent=True)
	except:
		efc = 0
	max_force = np.max(np.linalg.norm(a.get_forces(), axis=1))
	parprint("Energy before relaxation: {}, force consistent: {}, max_force: {}".format(e, efc, max_force))
	if max_force > 10.0:
		return np.inf

	if args.relax != 'No':
		fixed_zlevel = a.info.get('fixed_zlevel', None)
		if fixed_zlevel is not None:
			fixed_zlevel = float(fixed_zlevel)
			fix_mask = a.positions[:,2] < fixed_zlevel
			constr = FixAtoms(mask = fix_mask)
			a.set_constraint(constr)
		else:
			fix_mask = np.full(len(a), False)
		
		# Perturb to avoid saddle point stucking
		if args.perturbation > 0.:
			rnd_shifts = np.random.normal(scale = 1., size=(len(a), 3))
			rnd_shifts *= args.perturbation * np.broadcast_to((1. - fix_mask)/np.linalg.norm(rnd_shifts, axis=1), (3, len(a))).T
			rnd_shifts = MPI.COMM_WORLD.bcast(rnd_shifts, root=0)
			a.positions += rnd_shifts
		
		trajfn = args.log_dir + args.label+".traj"
		opt = relax(a, trajectory=trajfn)
		opt.run(fmax=args.fmax)
		e = a.get_potential_energy()
		try:
			efc = a.get_potential_energy(force_consistent=True)
		except:
			efc = 0
		parprint("Energy after relaxation: {}, (force consistent: {})".format(e, efc))
		try:
			traj = read(trajfn, index=':')
			write(args.log_dir + args.label + ".traj.xyz", traj, 'extxyz')
		except:
			pass
	
	a.set_calculator(None) # clean memory
	return e, a.copy()

